#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Created on Wed Jan 10 10:03:29 2018

@author: Kévin Giulietti and Guillaume Lavisse
"""

import numpy as np
import os
import cv2
import sys
import datetime
import matplotlib.pyplot as plt
from tkinter import Tk,Label,Button
from tkinter.filedialog import askdirectory,askopenfilename
from scipy.ndimage.morphology import binary_dilation, binary_erosion, binary_fill_holes
from skimage.filters import threshold_otsu
from skimage.io import imsave, imread_collection
from skimage.morphology import disk, remove_small_objects,binary_closing
from skimage.measure import regionprops, label
from skimage import *

###Default values
start_main=True 
current_directory=False     # If Atols is executed in current directory through command line
selected_threshold = None 
selected_mode = "F"         #Filamentous or Agregats
selected_lambda = 4 #default
list_option=["-c","-f","-a","-l","-t","-m","-e","-h"] #c : current direcoty - f:filamentous / a : agregat - l : lambda - t : threshold -m: min size - -h : help
single_image = None
args=[]
not_option=[]
min_size=10
ext=".tif"
def check_options():
    if len(sys.argv) > 1 :
        global start_main
        global current_directory
        global selected_threshold
        global selected_mode  
        global selected_lambda
        global list_option
        global single_image
        global args
        global not_option
        global min_size
        global ext
        for value in sys.argv[1:] :          
            try :
                if value.lower() in list_option :
                    args.append(value.lower())      #value here is one of the option
                elif value.endswith('.tif') or value.endswith('.jpg') or value.endswith('.png') :
                    try :
                        imread(os.getcwd() + "/" +value)  #If Atols successfully read the value, it's an image, ese it's just the extension type for file
                        single_image = value            #value here is the name of the image
                    except:
                        continue
                elif value.lower()[:2] == "-l" and float(value.lower()[2:]):
                    selected_lambda = float(value.lower()[2:])
                    
                elif value.lower()[:2] == "-t" and int(value.lower()[2:]):
                    selected_threshold = int(value.lower()[2:])

                elif value.lower()[:2] == "-m" and int(value.lower()[2:]):
                    min_size = int(value.lower()[2:])  
                    
                elif value.lower()[:2] == "-e" and len(value) > 2:
                    ext = value.lower()[2:]                 
                else:
                    not_option.append(value)        #Unknown option
            except:
                not_option.append(value)        #Unknown option
                 
        if "-h" in args:
            start_main = False
            print("Atols is a Python script allowing to detect features on images using a contrast scoring and adaptating threshold.\nList of options :\n")    
            print("-c : Treat all images in current directory. Otherwise, a window will appear, letting the user the choice between treating a whole directory or a single image.\n\n")
            print("-f : By default. General score mode, tends to find a compromise between small compacted and filamentous strucures\n ")
            print("-a : Score mode favors small compacted areas.\n\n ")
            print("-l : Lambda. Default value : 4 (Max). Used in score mode f, the higher is lambda the mode elongated and big structures are favored.\n\tExample : atols.py -l1.5\n")
            print("-t : Threshold to not detect features under a certain grayscale.\nExample : atols.py -t40\n")
            print("-m : Minimal size. Default value : 10. Remove structures smaller than x pixels.\n\tExample : atols.py -m15\n\n")
            print("-e : Default tif. Extension of images : png, tif, jpg.\n\tatols.py -ejpg.\n\n")
            print("If a path to a image is given, only this image will be treated.\n\tExample : atols.py /MyFolder/foo.tif")
        elif not_option: #Checking if all options are availables
                start_main = False
                print("Unknown option(s) : " + str(not_option))
                print("Type -h to display help.")
                
        #Checking if execute in current directory        
        if "-c" in args :
            current_directory = True
        #Check mode filamentous or agregats
        if "-a" in args:
            selected_mode="A"
        else:               #-f by default
            selected_mode="F"



def find_files(path, extension = ext):  # permet de trouver le type de fichiers voulu
    files = []
    for file in os.listdir(path):
        if file.endswith(extension):
            files.append(file)
        files.sort()
    return files

def find_best_pic(images):  # permet de trouver la meilleure image dans un stack

    if len(np.shape(images)) == 4:  # Si on a des stacks d'images RGB
        p = []
        for pic in range(len(images)):  # 1ère
            grey = cv2.cvtColor(images[pic], cv2.COLOR_BGR2GRAY)
            p += [threshold_otsu(grey)]  # On récup le seuil de chaque image
        p.sort()
        median_ots = np.median(p)  # Et on ne prend aue le median
    elif len(np.shape(images)) == 3 and np.shape(images)[-1] != 3:  # si c'est stacks d'images grises
        p = []

        for pic in range(len(images)):
            if images[pic].any():
                p += [threshold_otsu(images[pic])]
        p.sort()
        median_ots = np.median(p)

    else:
        grey = cv2.cvtColor(images, cv2.COLOR_BGR2GRAY)
        median_ots = threshold_otsu(grey)

    return median_ots


def labelisation(image):
    image_labelled = label(image)
    colorimage = np.zeros((image_labelled.shape[0],image_labelled.shape[1], 3), dtype = 'uint16')
    colorimage[:, :, 0] = image_labelled / 65536
    colorimage[:, :, 1] = image_labelled / 256
    colorimage[:, :, 2] = image_labelled % 256
    return colorimage


def scoring(image,mask_image,score_mode="F"):
    global selected_lambda
    mask_inner = mask_image.astype("bool") ^ binary_erosion(mask_image).astype("bool")                             # Couronne intérieure
    mask_border = binary_dilation(mask_image.astype("uint8"), disk(1)).astype("bool") ^ mask_image.astype("bool")  # Couronne extérieure
    region = image * mask_inner.astype("bool")
    border = image * mask_border.astype("bool")
    mean_i = np.sum(region) / np.sum(mask_inner)
    mean_e = np.sum(border) / np.sum(mask_border)
    if mean_e > 0:
        (mean_i - mean_e) + selected_lambda * (np.sum(mask_inner)) / np.sqrt(4 * np.pi * np.sum(mask_image))            
        
        if score_mode == 'F':
            score = (mean_i - mean_e) + selected_lambda * (np.sum(mask_inner)) / np.sqrt(4 * np.pi * np.sum(mask_image))
        elif score_mode == 'A':
            score = (mean_i - mean_e) / np.sqrt(np.var(region[region.nonzero()]) + np.var(
                    border[border.nonzero()]))  # On récupère uniquement la région en excluant le fond
        return score
    else:
        return 0
     
def demoat(image, threshold = None, mode = 'F'):
    try:
        imgrey = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    except:
        imgrey = np.copy(image)
           
    if threshold is None :
        min_value = int(threshold_otsu(imgrey)*0.8) #In order to not start at 0, otherwise the image will be full of noise
    else:
        min_value = threshold
    max_value=np.amax(imgrey)
    demoat_start=False
    for seuil in range(max_value - 5, min_value + 3, -3):       #We're not really beginning at top intensity + step of 3 to make it faster          
        if demoat_start==False : 
            lvl_n = imgrey>seuil                                #Binary mask at level n
            lvl_n = remove_small_objects(lvl_n,min_size)           
            
            if lvl_n.any() and binary_erosion(lvl_n).any() :    #check if not empty, in case of RSM or BE did it
                lvl_n_labelled   = label(lvl_n)
                mask_n = np.zeros_like(lvl_n)                   #Matrix which will be filled with starting areas 
                
                for n_regions in range(1,np.amax(lvl_n_labelled)+1)       :
                    if binary_erosion(lvl_n_labelled == n_regions).any():   #Add regions if they're big enough
                       mask_n += lvl_n_labelled == n_regions
                lvl_n=np.copy(mask_n)
                del(mask_n)
                demoat_start = True
        else:
            lvl_np1 = imgrey > seuil                             #Binary mask at level n   , lvl_n will be now be the n-1 mask
            lvl_np1 = remove_small_objects(lvl_np1,min_size)
            
            lvl_np1_labelled = label(lvl_np1)
            
            mask_np1 = np.zeros_like(lvl_np1)
            
            for np1_regions in range(1,np.amax(lvl_np1_labelled)+1)            :
                if binary_erosion(lvl_np1_labelled == np1_regions).any():
                   mask_np1 += lvl_np1_labelled == np1_regions
            lvl_np1 = np.copy(mask_np1)
            del(mask_np1)
            
            lvl_np1_labelled = label(lvl_np1)
            
            connexed_comp_n   =   regionprops(lvl_n_labelled)       #Features for each area in lvl_n
            connexed_comp_np1 =   regionprops(lvl_np1_labelled)     #Features for each area in lvl_np1
             
            number_connexed_comp_n   = len(connexed_comp_n)         #Number of areas in lvl_n
            number_connexed_comp_np1 = len(connexed_comp_np1)       #Number of areasin lvl_np1
            
            del(connexed_comp_n)
            del(connexed_comp_np1)
            
            np1_contains_n = [ [] for idx in range(number_connexed_comp_np1+1) ]
            #+1 will create a supplementary empty list. the 1st list will always
            #be empty since we're going to create other lists in which the index position
            #of each list will represent a unique area of lvl_np1. And each of these lists
            #will contain the regions of lvl_n included in the area corresponing in lvl_n, and so 
            #the position of the list.
            
            for label_in_n in range(1,number_connexed_comp_n+1) :
                [row,col]=(lvl_n_labelled==label_in_n).nonzero()    #Coordinates of the region number "label_in_n"
                label_np1_value=lvl_np1_labelled[row[0],col[0]]     #only need 1 pair of coordinates to get label value
                np1_contains_n[label_np1_value].append(label_in_n)  #At the position  "label_np1_value" in list, we add the "label_in_n" region
                #Example : the labelled region #2 of np1 (which is also he position #2 in the list) will contains the regions included in n
                
            for np1_CC in range(1,number_connexed_comp_np1+1) : #()
                score_n=0

                nb_n_CC_in=len(np1_contains_n[np1_CC])                             #number ot Connexed Components (CC) n included in np1
                for n_CC in range(nb_n_CC_in):
                    mask_CC_n = lvl_n_labelled ==   np1_contains_n[np1_CC][n_CC]    #Mask of one of the CC in n included in the CC of np1
                    score_n+=scoring(imgrey,mask_CC_n,mode)              #Total score of regions included in np1
                mask_CC_np1   = lvl_np1_labelled==np1_CC
                score_np1 = scoring(imgrey,mask_CC_np1,mode)
                if score_np1 is None or score_np1 > score_n :       #None score appeared sometimes for unknown, it strangely happens when there is a new area, so we had it anyway
                    lvl_n = (lvl_n + (lvl_np1_labelled == np1_CC)).astype("bool") #We had them in the new lvl_n  matrix
        
    final = lvl_n.astype("bool")
    return final
    
def Atols_execution(path_to_image,threshold=None,mode="F"):
    file = imread_collection(path_to_image)
    
    if threshold is None:
        threshold = int(find_best_pic(file))
        
    im_base = file[0]
    final = demoat(im_base, threshold, mode)
    
    now = datetime.datetime.now()
    print(str(now))
    
    final_labeled = labelisation(final)
    basename,filename=os.path.split(path_to_image)
    path_result="/".join([basename,"Atols_results"])
    if not os.path.exists(path_result):   
        os.mkdir(path_result)
        
    imsave(("/Mask_").join([path_result, filename]), img_as_ubyte(final))
    imsave(("/Labeled_").join([path_result, filename]), img_as_uint(final_labeled))  
#    #Otsu mask in order to compare with Atols Binarization
#    try:
#        gris=cv2.cvtColor(im_base, cv2.COLOR_BGR2GRAY)
#    except :
#        gris=im_base
#    otsu = gris > threshold_otsu(gris)
#    imsave(("/Otsu_").join([path_result, filename]), img_as_ubyte(otsu))
    

    
    
def directory():
    fenetre.destroy()
    directoryname_window = Tk()
    directoryname_window.withdraw()
    path = askdirectory(title = "Choose a directory containing images")
    if path !=():
        my_images = [name for name in find_files(path) if name.endswith(ext)]
        total_images = len(my_images)
        for image_position in range(total_images):
            print("\nFile " + str(image_position + 1) + "/" + str(total_images) + " : " + my_images[image_position])
            path_image = "/".join([path, my_images[image_position]])
            # la position de l'image donne le label de l'image qui correspond son emplacement dans les sous_dossiers
            Atols_execution(path_image,selected_threshold,selected_mode)
        directoryname_window.destroy()

    else:
        directoryname_window.destroy()

        
def file():
    fenetre.destroy()
    filename_window=Tk()
    filename_window.withdraw()
    path_image=askopenfilename(filetypes=[("Choose the image to binarize","*.tif *.jpg *.png")])
    if path_image !=():
        print("\nFile " + os.path.split(path_image)[1] ) # = filename
        Atols_execution(path_image,selected_threshold,selected_mode)
        filename_window.destroy()
    else:
        filename_window.destroy()
    
########################################## MAIN ##########################################

if __name__ ==  "__main__" and start_main :
    check_options()
    if start_main:
        print(datetime.datetime.now())
        if single_image : # or (current_directory and single_image):
            path_image=os.getcwd() + "/" + single_image
            Atols_execution(path_image,selected_threshold,selected_mode)
            
        elif current_directory :
            
            my_path=os.getcwd()
            my_images = [name for name in find_files(my_path,ext)]
            n = len(my_images)
            for number_images in range(n):
                print("\nFile execution ",number_images + 1,"/",n,": " + my_images[number_images])
                path_image = "/".join([my_path, my_images[number_images]])
                Atols_execution(path_image,selected_threshold,selected_mode)
        else:
            fenetre=Tk()
            fenetre.title("Select type of operation")
            lab=Label(fenetre,text="Choose between a single image or a directory for your pictures")
            lab.pack(padx=5,pady=5)
            Bout=Button(fenetre,text="File",command = file)
            Bout.pack()
            Bout1=Button(fenetre,text="Directory",command = directory)
            Bout1.pack()
            fenetre.mainloop()
